import os
import datetime
from wed.utils import DataType, Service
from wed import settings


class DatasetPath:
    """
    Describe the location of a Dataset.
    A dataset is a set of one file per day
    """
    def __init__(self, bucket: str, local_cache_path: str,
                 type: DataType, service: Service):
        self.bucket = bucket
        self.local_cache_path = local_cache_path
        self.type = type
        self.service = service

    @property
    def prefix(self):
        return settings.get_prefix(self.type, self.service)

    @property
    def local_path(self):
        return os.path.join(self.local_cache_path, self.prefix)


class DatafilePath:
    """
    Describe the location of a Datafile.
    A datafile is a single file either remote or locally hosted
    """
    def __init__(self, dataset_path: DatasetPath, time: datetime.datetime):
        self.path = dataset_path
        self.__filename = time.isoformat().split('T')[0] + ".grb"

    @property
    def local_path(self):
        return os.path.join(self.path.local_path, self.__filename)

    @property
    def bucket(self):
        return self.path.bucket

    @property
    def prefix(self):
        return os.path.join(self.path.prefix, self.__filename)



import unittest
import math
import datetime
import shutil
import logging
logging.basicConfig(level=logging.INFO)

import wed

import functools

@functools.lru_cache(maxsize=10)
def test(i: int):
    print("in test")
    return 0

class HighLevelTest(unittest.TestCase):
    def test_basic(self):
        coord = wed.Coordinate(0, 0)

        wed.Settings().latitude_precision = 0.5
        wed.Settings().longitude_precision = 0.5
        wed.Settings().time_precision = datetime.timedelta(hours=6)
        for i in range(1000):
            # vary the time and coordinate according to i so that we can check the caching behavior
            t = datetime.datetime(2017, 5, 12, math.floor(i/60), i % 60)
            coord = wed.Coordinate(i/100, i/100)
            wed.read(wed.DataType.WIND, t, coord)

class DatabaseTest(unittest.TestCase):
    def test_basic(self):
        t = datetime.datetime(2017, 5, 12, 20, 4)
        print(wed.read(wed.DataType.WIND, t, wed.Coordinate(0, 0)))
        print(wed.read(wed.DataType.WIND, t, wed.Coordinate(12, 150)))

    def test_negative(self):
        c1 = wed.round_coordinate(wed.Coordinate(0, -90), wed.DataType.WIND)
        c2 = wed.round_coordinate(wed.Coordinate(0, 270), wed.DataType.WIND)
        self.assertEqual(c1.latitude, c2.latitude)
        self.assertEqual(c1.longitude, c2.longitude)

    def test_range(self):
        t = datetime.datetime(2019, 5, 12, 20, 4)
        range = wed.get_time_range(wed.DataType.WIND)
        def is_in_range(time: datetime.datetime) -> bool:
            return range['min'] <= time <= range['max']
        self.assertFalse(is_in_range(datetime.datetime(2019, 5, 12, 20, 4)))
        self.assertTrue(is_in_range(datetime.datetime(2017, 5, 12, 20, 4)))

    def test_basic_wave(self):
        t = datetime.datetime(2017, 5, 12, 20, 4)
        print(wed.read(wed.DataType.WAVE, t, wed.Coordinate(0, 0)))
        print(wed.read(wed.DataType.WAVE, t, wed.Coordinate(12, 150)))



import boto3
import os
import errno

from wed.path import DatafilePath


def mkdirs(path):
    """
    Build the directory if it does not exist yet
    """
    try:
        os.makedirs(path)
    except OSError as err:
        if err.errno != errno.EEXIST or not os.path.isdir(path):
            raise

def download(path: DatafilePath):
    """
    Download the Datafile to the cache directory
    """
    s3 = boto3.client('s3')
    mkdirs(os.path.dirname(path.local_path))
    s3.download_file(path.bucket, path.prefix, path.local_path)



import os
import datetime
import functools
import logging

import cfgrib
import xarray
import numpy as np

from wed import settings
from wed.path import DatasetPath, DatafilePath
from wed.utils import Coordinate, DataType, Service, round_to_precision
from wed.network import download


## ---- Cached functions is used to speed up accesses
# Use the LRU cache property to avoid reopenning the dataset each time
@functools.lru_cache(maxsize=100)
def get_dataset(bucket:str,
                service: Service,
                local_cache_path: str,
                type: DataType,
                timestamp: float) -> xarray.Dataset:
    time = datetime.datetime.fromtimestamp(timestamp)
    dataset_path = DatasetPath(bucket, local_cache_path, type, service)
    file_path = DatafilePath(dataset_path, time)
    if not os.path.exists(file_path.local_path):
        logging.info("Datafile at " + file_path.local_path + " not found. Downloading...")
        download(file_path)
        logging.info("Download successful")
    return cfgrib.open_dataset(file_path.local_path)

# To use the lru_cache effectively we need to keep it to simple types (no datetime of coordinate)
@functools.lru_cache(maxsize=10000)
def get_data(bucket:str,               # The s3 bucket where the database can be found
             service: Service,         # The service to be used
             local_cache_path: str,    # The place were datafiles were downloaded
             type: DataType,           # The type of data to download (wind / wave ...)
             timestamp: float,         # The date of the desired data as timestamp
             latitude: float,
             longitude: float) -> dict:
    time = datetime.datetime.fromtimestamp(timestamp)
    dataset = get_dataset(bucket, service, local_cache_path, type, timestamp)
    value = dataset.interp({
        'latitude': latitude,
        'longitude': longitude,
        'time': np.datetime64(time)
    }, method='linear')

    retval = {}
    for var in value.data_vars:
        retval[var] = float(value[var].values)
    return retval

def read(type: DataType, time: datetime.datetime, coordinate: Coordinate) -> dict:
    # Round the time
    timestamp = time.timestamp()
    time_precision = settings.Settings().time_precision
    if time_precision is not None:
        timestamp = round_to_precision(timestamp, time_precision.total_seconds())

    return get_data(settings.Settings().bucket,
                    settings.Settings().services[type],
                    settings.Settings().local_cache_path,
                    type,
                    timestamp,
                    round_to_precision(coordinate.latitude, settings.Settings().latitude_precision),
                    round_to_precision(coordinate.longitude, settings.Settings().longitude_precision))



from enum import Enum, auto
from wed import settings


class DataType(Enum):
    WIND = auto()
    WAVE = auto()
    CURRENT = auto()

    def __str__(self):
        return self.name.lower()

class Service(Enum):
    ECMWF = auto()

    def __str__(self):
        return self.name.lower()

class Coordinate:
    def __init__(self, latitude: float, longitude: float):
        self.latitude = latitude
        self.longitude = longitude

def round_to_precision(value: float, precision: float) -> float:
    if precision is None:
        return value
    rest = value % precision
    value -= rest
    if rest >= precision / 2:
        value += precision
    return value

def round_coordinate(coordinate: Coordinate, type: DataType) -> Coordinate:
    """Update the coordinate considering the range"""
    range = settings.get_coordinate_range(type, settings.Settings().services[type])

    longitude = coordinate.longitude
    while longitude < range["longitude"]["min"]:
        longitude += 360
    while longitude > range["longitude"]["max"]:
        longitude -= 360
    coordinate.longitude = longitude
    return coordinate



import os
import json
import datetime
from wed.utils import DataType, Service

# ---- Constants
## List the remote path to the datasets.
# You Need to update it when adding new stuff
settings_path = os.path.join(os.path.dirname(os.path.realpath(__file__)), "settings.json")
with open(settings_path) as settings_file:
    SETTINGS = json.load(settings_file)

# Default values
DEFAULT_BUCKET = "dice-weather-data"
DEFAULT_CACHE_PATH = os.environ['HOME'] + "/.wed-cache"
DEFAULT_SERVICES = {
    DataType.WIND: Service.ECMWF,
    DataType.WAVE: Service.ECMWF
}
# ---- /Constants

class Settings:
    """Singleton used to store and modify wed settings"""
    class __Settings:
        def __init__(self):
            self.bucket = DEFAULT_BUCKET
            self.services = DEFAULT_SERVICES
            self.local_cache_path = DEFAULT_CACHE_PATH
            self.latitude_precision = None
            self.longitude_precision = None
            self.time_precision = None

    instance = None
    def __init__(self):
        if not Settings.instance:
            Settings.instance = Settings.__Settings()

    def __getattr__(self, name):
        return getattr(self.instance, name)
    def __setattr__(self, name, value):
        setattr(self.instance, name, value)


def get_prefix(type: DataType, service: Service):
    return SETTINGS[str(type)][str(service)]['prefix']

def get_time_range(type: DataType):
    service = Settings().services[type]
    range = SETTINGS[str(type)][str(service)]['range']['date']
    return {
        'min': datetime.datetime.strptime(range["min"], '%Y-%m-%d'),
        'max': datetime.datetime.strptime(range["max"], '%Y-%m-%d')
    }

def get_coordinate_range(type: DataType, service: Service):
    lat_range = SETTINGS[str(type)][str(service)]['range']['latitude']
    lon_range = SETTINGS[str(type)][str(service)]['range']['longitude']
    return {
        "latitude": lat_range,
        "longitude": lon_range
    }



import os
import setuptools

# parse the requirements.txt
current_path = os.path.dirname(os.path.realpath(__file__))
requirements = []
with open(os.path.join(current_path, "requirements.txt")) as reqfile:
    for requirement in reqfile.read().splitlines():
        requirements.append(requirement)

setuptools.setup(
    name="pywed",
    version="0.0.5",
    author="Guillaume Vincke",
    author_email="guillaume.vincke@dice-engineering.com",
    description="PYthon client to d-ice WEthear Database",
    long_description="If you don't have a valid access key to the database, you won't be able to use this",
    url="https://d-ice.gitlab.host/common/python/pywed",
    packages=['wed'],
    install_requires=requirements,
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    include_package_data=True
)
